from python_sdk_remote.utilities import get_environment_name
from url_remote.environment_name_enum import EnvironmentName

from .to_sql_interface import ToSQLInterface


def validate_select_table_name(database_object_name: str) -> None:
    if (get_environment_name() not in (EnvironmentName.DVLP1.value, EnvironmentName.PROD1.value)
            and not database_object_name.endswith("_view")):
        raise Exception(
            f"View name must end with '_view' in this environment (got {database_object_name})")


def validate_none_select_table_name(database_object_name: str) -> None:
    if (get_environment_name() not in (EnvironmentName.DVLP1.value, EnvironmentName.PROD1.value)
            and not database_object_name.endswith("_table")):
        raise Exception(
            f"Table name must end with '_table' in this environment  (got {database_object_name})")


def process_insert_data_json(data_json: dict or None) -> tuple[str, str, dict]:
    if not data_json:
        return '', '', {}

    columns = []
    values = []

    for key, value in data_json.items():
        columns.append(f"`{key}`")
        if isinstance(value, ToSQLInterface):
            values.append(value.to_sql())
        else:
            values.append('%s')

    filtered_data_json = {key: value for key, value in data_json.items() if
                          not isinstance(value, ToSQLInterface)}
    return ','.join(columns), ','.join(values), filtered_data_json


# Please add typing and example of input-output as docstring if possible.
def process_update_data_json(data_json: dict or None) -> tuple[str, dict]:
    if not data_json:
        return '', {}

    set_values = []
    for key, value in data_json.items():
        if isinstance(value, ToSQLInterface):
            set_values.append(f"`{key}`={value.to_sql()}")
        else:
            set_values.append(f"`{key}`=%s")

    filtered_data_json = {key: value for key, value in data_json.items() if
                          not isinstance(value, ToSQLInterface)}
    # + "," because we add updated_timestamp in the update query
    return ', '.join(set_values) + ",", filtered_data_json



from .to_sql_interface import ToSQLInterface


class Point(ToSQLInterface):
    def __init__(self, longitude: float, latitude: float) -> None:
        self.longitude = longitude
        self.latitude = latitude

    def to_sql(self) -> str:
        return f"POINT ({self.longitude}, {self.latitude})"

    @staticmethod
    def create_select_stmt(column_name: str) -> str:
        return f"ST_X({column_name}), ST_Y({column_name})"

    def __eq__(self, other_point: 'Point') -> bool:
        return (isinstance(other_point, Point) and
                self.longitude == other_point.longitude and
                self.latitude == other_point.latitude)

    def __repr__(self) -> str:
        return f"Point(longitude={self.longitude}, latitude={self.latitude})"



from abc import ABC, abstractmethod


# TODO: should we use SqlLiteral instead of ToSQLInterface and all its subclasses?
class ToSQLInterface(ABC):
    """
    An interface for objects that represent structures to be
    inserted into a database.

    Subclasses must implement the `to_sql` method, which should return a string
    representing the SQL representation of the structure.

    Example:
    --------
    class Point(ToSQLInterface):
        def __init__(self, longitude, latitude):
            self.longitude = longitude
            self.latitude = latitude

        def to_sql(self):
            return f"POINT({self.longitude}, {self.latitude})"
    """

    @abstractmethod
    def to_sql(self) -> str:
        pass


class Now(ToSQLInterface):
    """
    Represents the current time in SQL.
    """

    def to_sql(self):
        return "NOW()"


class CurrentDate(ToSQLInterface):
    """
    Represents the current time in SQL.
    """

    def to_sql(self):
        return "CURDATE()"


class TimeStampDiff(ToSQLInterface):
    """
    Represents the current time in SQL.
    """

    def __init__(self, unit, start, end):
        allowed_units = ("YEAR", "QUARTER", "MONTH", "WEEK", "DAY", "HOUR", "MINUTE", "SECOND", "FRAC_SECOND")
        if unit not in allowed_units:
            raise ValueError(f"unit must be one of {allowed_units}")
        self.unit = unit
        self.start = start
        self.end = end

    def to_sql(self):
        return f"TIMESTAMPDIFF({self.unit}, {self.start}, {self.end})"


class Count(ToSQLInterface):
    """Represents a COUNT() function in SQL."""

    def __init__(self, column_name="*"):
        self.column_name = column_name

    def to_sql(self):
        return f"COUNT({self.column_name})"



import re
from typing import Any
from typing import Optional

from language_remote.lang_code import LangCode
from logger_local.MetaLogger import MetaLogger
from user_context_remote.user_context import UserContext

from .constants import DEFAULT_SQL_SELECT_LIMIT, LOGGER_CRUD_ML_CODE_OBJECT
from .generic_crud import GenericCRUD
from .utils import validate_select_table_name

IS_MAIN_COLUMN_NAME = "is_main"


class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=LOGGER_CRUD_ML_CODE_OBJECT):
    """A class that provides generic CRUD functionality for tables with multi-language support."""

    def __init__(self, default_schema_name: str, default_table_name: str = None,
                 default_id_column_name: str = None,
                 is_main_column_name: str = IS_MAIN_COLUMN_NAME,
                 is_test_data: bool = False) -> None:
        """Initializes the GenericCRUDML class. If connection is not provided,
        a new connection will be created."""
        if default_table_name is not None:
            self.default_ml_table_name = self._create_ml_table_name(default_table_name)
            self.default_view_table_name = self._get_ml_view_name(default_table_name)
        else:
            self.default_ml_table_name = None
            self.default_view_table_name = None
        self.logger.info("GenericCRUDML.__init__", object={"default_ml_table_name": self.default_ml_table_name,
                                                           "default_view_table_name": self.default_view_table_name})
        GenericCRUD.__init__(self, default_schema_name=default_schema_name,
                             default_table_name=default_table_name,
                             default_view_table_name=self.default_view_table_name,
                             default_id_column_name=default_id_column_name,
                             is_test_data=is_test_data)
        self.user_context = UserContext()
        # TODO: we may have to change it to a list when we will have multiple is_main columns
        self.is_main_column_name = is_main_column_name

    def sql_in_list_by_entity_list_id(self, schema_name: str, entity_name: str, entity_list_id: int) -> str:
        """Example: select group_id from group.group_list_member_table WHERE group_list_id=1"""

        old_schema_name = self.schema_name
        self.set_schema(schema_name=schema_name)
        ids = self.select_multi_dict_by_id(view_table_name=f"{entity_name}_list_member_view",
                                           select_clause_value=f"{entity_name}_id",
                                           id_column_name=f"{entity_name}_list_id",
                                           id_column_value=entity_list_id)
        if ids:
            result = f" IN ({','.join([str(_id[f'{entity_name}_id']) for _id in ids])})"
        else:
            result = " = TRUE"
        self.set_schema(schema_name=old_schema_name)

        return result

    def add_value(self, data_ml_json: dict, table_id: int = None, lang_code: LangCode = None,
                  is_main: int = 0, data_json: dict = None, table_name: str = None,
                  ml_table_name: str = None, schema_name: str = None) -> Optional[tuple]:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_json=data_ml_json)
        old_schema_name = self.schema_name
        self.schema_name = schema_name or self.schema_name
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name

        old_cursor = self.cursor
        try:
            cursor = self.connection.cursor()
            self.cursor = cursor

            # if this is the first insert of this data, is_main should be 1
            if table_id is None:
                is_main = 1
            elif is_main == 1:
                self._update_old_main_value_to_zero(table_id, table_name)

            # id is the id value of the row in the table_name table
            table_id = table_id or self.insert(data_json=data_json, ignore_duplicate=True,
                                               table_name=table_name, commit_changes=False)
            if table_id is None:  # TODO: delete
                self.logger.error("Error inserting data_json", object={"data_json": str(data_json)})
                self.connection.rollback()
                return None

            if data_ml_json is None or data_ml_json == {}:
                self.connection.commit()
                return table_id, None

            id_column_name = re.sub(r'(_table)$', '_id', table_name)
            data_ml_json[id_column_name] = table_id
            data_ml_json["lang_code"] = lang_code_str
            data_ml_json[self.is_main_column_name] = is_main

            # ml_id is the id value of the row in the ml_table_name table
            ml_table_id = self.insert(data_json=data_ml_json, table_name=ml_table_name, ignore_duplicate=True,
                                      commit_changes=False)

            self.connection.commit()
        except Exception as e:
            self.connection.rollback()
            raise e
        finally:
            self.cursor = old_cursor
            self.schema_name = old_schema_name
            cursor.close()
        return table_id, ml_table_id

    def add_value_if_not_exist(self, data_ml_json: dict, table_id: int = None, lang_code: LangCode = None,
                               is_main: int = 0, data_json: dict = None, table_name: str = None,
                               ml_table_name: str = None, schema_name: str = None,
                               order_by: str = None) -> tuple:
        self.schema_name = schema_name or self.schema_name
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        id_column_name = GenericCRUD.generate_id_column_name(table_name)
        ml_id_column_name = GenericCRUD.generate_id_column_name(ml_table_name)

        # check if the row exists in the ml_table
        name = data_ml_json.get("title") or data_ml_json.get("name")
        if not name:
            raise ValueError("name or title is required for upsert_value")
        table_id = table_id or self.get_id_by_name(table_name=table_name, ml_table_name=ml_table_name, name=name,
                                                   lang_code=lang_code,
                                                   id_column_name=id_column_name, order_by=order_by)
        ml_table_id = self.get_ml_id_by_name(ml_table_name=ml_table_name, name=name, lang_code=lang_code,
                                             id_column_name=ml_id_column_name, order_by=order_by)
        if not table_id:
            return self.add_value(data_ml_json=data_ml_json, table_id=table_id, lang_code=lang_code, is_main=is_main,
                                  data_json=data_json, table_name=table_name, ml_table_name=ml_table_name)
        else:
            return table_id, ml_table_id

    def update_value_by_id(self, data_ml_json: dict, ml_table_id: int, table_id: int,
                           lang_code: LangCode = None, is_main: int = 0, data_json: dict = None,
                           table_name: str = None, ml_table_name: str = None,
                           schema_name: str = None, limit: int = DEFAULT_SQL_SELECT_LIMIT,
                           order_by: str = None) -> tuple:
        old_schema_name = self.schema_name
        self.schema_name = schema_name or self.schema_name
        lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_json=data_ml_json)
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        id_column_name = GenericCRUD.generate_id_column_name(table_name)

        if table_id is None:
            raise ValueError("table_id is required for update_value")
        if ml_table_id is None:
            raise ValueError("ml_table_id is required for update_value")

        old_cursor = self.cursor
        try:
            cursor = self.connection.cursor()
            self.cursor = cursor

            if is_main == 1:
                self._update_old_main_value_to_zero(table_id, table_name)
            if data_json:
                self.update_by_id(data_json=data_json, table_name=table_name, id_column_name=id_column_name,
                                  id_column_value=table_id, limit=limit, order_by=order_by, commit_changes=False)

            data_ml_json[id_column_name] = table_id
            data_ml_json["lang_code"] = lang_code_str
            data_ml_json[self.is_main_column_name] = is_main
            id_column_name = GenericCRUD.generate_id_column_name(ml_table_name)

            self.update_by_id(data_json=data_ml_json, table_name=ml_table_name, id_column_name=id_column_name,
                              id_column_value=ml_table_id, limit=limit, order_by=order_by, commit_changes=False)

            self.connection.commit()
        except Exception as e:
            self.connection.rollback()
            raise e
        finally:
            self.cursor = old_cursor
            self.schema_name = old_schema_name
            cursor.close()
        return table_id, ml_table_id

    def upsert_value(self, data_ml_json: dict, table_id: int = None, 
                     data_json_compare: dict = None, lang_code: LangCode = None,
                     is_main: int = 0, data_json: dict = None, table_name: str = None,
                     ml_table_name: str = None, schema_name: str = None,
                     limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> tuple:
        self.schema_name = schema_name or self.schema_name
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        id_column_name = GenericCRUD.generate_id_column_name(table_name)
        ml_id_column_name = GenericCRUD.generate_id_column_name(ml_table_name)

        # check if the row exists in the ml_table
        name = data_ml_json.get("title") or data_ml_json.get("name")
        name_compare = name
        if not name:
            raise ValueError("name or title is required for upsert_value")
        if data_json_compare:
            name_compare = data_json_compare.get("title") or data_json_compare.get("name")

        table_id = table_id or self.get_id_by_name(table_name=table_name, ml_table_name=ml_table_name, name=name_compare,
                                                   lang_code=lang_code,
                                                   id_column_name=id_column_name, order_by=order_by)
        ml_table_id = self.get_ml_id_by_name(ml_table_name=ml_table_name, name=name_compare, lang_code=lang_code,
                                             id_column_name=ml_id_column_name, order_by=order_by)
        if not table_id:
            return self.add_value(data_ml_json=data_ml_json, table_id=table_id, lang_code=lang_code, is_main=is_main,
                                  data_json=data_json, table_name=table_name, ml_table_name=ml_table_name)
        else:
            return self.update_value_by_id(data_ml_json=data_ml_json, ml_table_id=ml_table_id, table_id=table_id,
                                           lang_code=lang_code, is_main=is_main, data_json=data_json,
                                           table_name=table_name,
                                           ml_table_name=ml_table_name, schema_name=schema_name,
                                           limit=limit, order_by=order_by)

    def upsert_value_with_abbreviations(self, data_ml_json: dict,  data_json_compare: dict = None,
                                        table_id: int = None, lang_code: LangCode = None,
                                        data_json: dict = None, table_name: str = None,
                                        ml_table_name: str = None, schema_name: str = None,
                                        limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> tuple:
        self.schema_name = schema_name or self.schema_name
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        id_column_name = GenericCRUD.generate_id_column_name(table_name)
        ml_id_column_name = GenericCRUD.generate_id_column_name(ml_table_name)

        # check if the row exists in the ml_table
        name_and_abbreviations = data_ml_json.get("title") or data_ml_json.get("name")
        if not name_and_abbreviations:
            raise ValueError("name or title is required for upsert_value")
        if data_json_compare:
            name_and_abbreviations_compare = data_json_compare.get("title") or data_json_compare.get("name")
        else:
            name_and_abbreviations_compare = name_and_abbreviations

        name = name_and_abbreviations.split(" (")[0]
        name_compare = name_and_abbreviations_compare.split(" (")[0]
        abbreviations = name_and_abbreviations.split(" (")[1][:-1]
        abbreviations_list = abbreviations.split(", ")
        data_ml_json_with_name = data_ml_json.copy()
        data_ml_json_with_name["title"] = name
        data_ml_json_with_abbreviation_list = []
        for abbreviation in abbreviations_list:
            data_ml_json_with_abbreviation = data_ml_json.copy()
            data_ml_json_with_abbreviation["title"] = abbreviation
            data_ml_json_with_abbreviation_list.append(data_ml_json_with_abbreviation)

        table_id = table_id or self.get_id_by_name(table_name=table_name, ml_table_name=ml_table_name, name=name_compare,
                                                   lang_code=lang_code,
                                                   id_column_name=id_column_name, order_by=order_by)
        ml_table_id = self.get_ml_id_by_name(ml_table_name=ml_table_name, name=name_compare, lang_code=lang_code,
                                             id_column_name=ml_id_column_name, order_by=order_by)
        ml_ids_list = []
        if not table_id:
            table_id, ml_table_id = self.add_value(data_ml_json=data_ml_json_with_name, table_id=table_id, lang_code=lang_code,
                                                   is_main=1, data_json=data_json, table_name=table_name,
                                                   ml_table_name=ml_table_name)
            ml_ids_list.append(ml_table_id)
            for data_ml_json_with_abbreviation in data_ml_json_with_abbreviation_list:
                table_id, ml_table_id = self.add_value(data_ml_json=data_ml_json_with_abbreviation, table_id=table_id,
                                                       lang_code=lang_code, is_main=0, data_json=data_json,
                                                       table_name=table_name, ml_table_name=ml_table_name)
                ml_ids_list.append(ml_table_id)
        else:
            ml_ids_list = self.get_ml_ids_by_id(table_id=table_id, ml_table_name=ml_table_name, lang_code=lang_code,
                                                id_column_name=id_column_name, ml_id_column_name=ml_id_column_name,
                                                order_by=order_by)
            ml_ids_list.remove(ml_table_id)
            table_id, main_ml_table_id = self.update_value_by_id(data_ml_json=data_ml_json_with_name, ml_table_id=ml_table_id,
                                                                 table_id=table_id,
                                                                 lang_code=lang_code, is_main=1, data_json=data_json,
                                                                 table_name=table_name,
                                                                 ml_table_name=ml_table_name, schema_name=schema_name,
                                                                 limit=limit, order_by=order_by)
            for index, data_ml_json_with_abbreviation in enumerate(data_ml_json_with_abbreviation_list):
                if index < len(ml_ids_list):
                    table_id, ml_table_id = self.update_value_by_id(data_ml_json=data_ml_json_with_abbreviation,
                                                                    ml_table_id=ml_ids_list[index], table_id=table_id,
                                                                    lang_code=lang_code, is_main=0, data_json=data_json,
                                                                    table_name=table_name,
                                                                    ml_table_name=ml_table_name, schema_name=schema_name,
                                                                    limit=limit, order_by=order_by)
                else:
                    table_id, ml_table_id = self.add_value(data_ml_json=data_ml_json_with_abbreviation, table_id=table_id,
                                                           lang_code=lang_code, is_main=0, data_json=data_json,
                                                           table_name=table_name, ml_table_name=ml_table_name)
                ml_ids_list.append(ml_table_id)
            ml_ids_list.append(main_ml_table_id)
        return table_id, ml_ids_list

    def get_values_tuple(self, table_id: int, lang_code: LangCode = None,
                         id_column_name: str = None, select_clause_value: str = "*",
                         order_by: str = None) -> tuple:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(
            self.default_table_name)
        result = self.select_one_tuple_by_where(where=f"{id_column_name}=%s AND lang_code=%s",
                                                params=(table_id, lang_code_str),
                                                select_clause_value=select_clause_value,
                                                order_by=order_by)

        return result

    def get_values_dict(self, table_id: int, lang_code: LangCode = None,
                        id_column_name: str = None,
                        order_by: str = None) -> dict:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(
            self.default_table_name)
        result = self.select_one_dict_by_where(where=f"{id_column_name}=%s AND lang_code=%s",
                                               params=(table_id, lang_code_str),
                                               order_by=order_by or f"{id_column_name} DESC")

        return result

    def get_values_dict_list(self, table_id: int, lang_code: LangCode = None,
                             id_column_name: str = None,
                             order_by: str = None) -> list:
        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(
            self.default_table_name)
        result = self.select_multi_dict_by_where(where=f"{id_column_name}=%s AND lang_code=%s",
                                                 params=(table_id, lang_code_str),
                                                 order_by=order_by)
        return result

    def get_main_values_tuple(self, table_id: int, id_column_name: str = None, select_clause_value: str = "*",
                              order_by: str = None) -> tuple:

        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(
            self.default_table_name)
        result = self.select_one_tuple_by_where(where=f"{id_column_name}=%s AND is_main=1",
                                                params=(table_id,),
                                                select_clause_value=select_clause_value,
                                                order_by=order_by or f"{id_column_name} DESC")

        return result

    def get_main_values_dict(self, table_id: int, id_column_name: str = None, select_clause_value: str = "*",
                             order_by: str = None) -> dict:

        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(
            self.default_table_name)
        result = self.select_one_dict_by_where(where=f"{id_column_name}=%s AND is_main=1",
                                               params=(table_id,),
                                               select_clause_value=select_clause_value,
                                               order_by=order_by)

        return result

    def get_id_by_name(self, name: str, table_name: str = None, ml_table_name: str = None, lang_code: LangCode = None,
                       id_column_name: str = None, order_by: str = None) -> Optional[int]:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        ml_view_name = self._get_view_name(ml_table_name)
        id_column_name = id_column_name or GenericCRUD.generate_id_column_name(table_name)
        if GenericCRUD.is_column_in_table(self, table_name=ml_table_name, column_name="title"):
            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
                                                    select_clause_value=id_column_name,
                                                    where="title=%s AND lang_code=%s",
                                                    params=(name, lang_code_str),
                                                    order_by=order_by)
        else:
            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
                                                    select_clause_value=id_column_name,
                                                    where="`name`=%s AND lang_code=%s",
                                                    params=(name, lang_code_str),
                                                    order_by=order_by)
        if result:
            return result[0]
        else:
            return None

    def get_ml_id_by_name(self, name: str, ml_table_name: str = None, lang_code: LangCode = None,
                          id_column_name: str = None, order_by: str = None) -> Optional[int]:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
        ml_table_name = ml_table_name or self.default_ml_table_name
        ml_view_name = self._get_view_name(ml_table_name)
        ml_id_column_name = id_column_name or GenericCRUD.generate_id_column_name(ml_table_name)
        if GenericCRUD.is_column_in_table(self, table_name=ml_table_name, column_name="title"):
            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
                                                    select_clause_value=ml_id_column_name,
                                                    where="title=%s AND lang_code=%s",
                                                    params=(name, lang_code_str),
                                                    order_by=order_by)
        else:
            result = self.select_one_tuple_by_where(view_table_name=ml_view_name,
                                                    select_clause_value=ml_id_column_name,
                                                    where="`name`=%s AND lang_code=%s",
                                                    params=(name, lang_code_str),
                                                    order_by=order_by)
        if result:
            return result[0]
        else:
            return None

    def get_ml_ids_by_id(self, table_id: int, ml_table_name: str = None, lang_code: LangCode = None,
                         id_column_name: str = None, ml_id_column_name: str = None, order_by: str = None) -> list:
        lang_code_str = None if lang_code is None else lang_code.value
        result = self.select_multi_tuple_by_where(view_table_name=self._get_view_name(ml_table_name),
                                                  select_clause_value=ml_id_column_name,
                                                  where=f"{id_column_name}=%s AND lang_code=%s",
                                                  params=(table_id, lang_code_str),
                                                  order_by=order_by)
        return [row[0] for row in result]

    def delete_by_name(self, name: str,
                       lang_code: LangCode = None) -> None:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
        if GenericCRUD.is_column_in_table(self, table_name=self.default_view_table_name, column_name="title"):
            self.delete_by_where(table_name=self.default_ml_table_name, where="title=%s AND lang_code=%s",
                                 params=(name, lang_code_str))
        else:
            self.delete_by_where(table_name=self.default_ml_table_name, where="`name`=%s AND lang_code=%s",
                                 params=(name, lang_code_str))

    @staticmethod
    def _create_ml_table_name(table_name: str) -> str:
        ml_table_name = re.sub(r'(_table)$', '_ml_table', table_name)
        return ml_table_name

    @staticmethod
    def _get_ml_view_name(table_name: str) -> str:
        view_name = re.sub(r'(_table)$', '_ml_view', table_name)
        return view_name

    # Change the old row with is_main=1 to is_main=0
    def _update_old_main_value_to_zero(self, table_id: int, table_name: str) -> None:

        data_ml_json = {self.is_main_column_name: 0}
        id_column_name = GenericCRUD.generate_id_column_name(table_name)
        self.update_by_where(table_name=self.default_ml_table_name,
                             data_json=data_ml_json,
                             where=f"{id_column_name}=%s AND is_main=1",
                             params=(table_id,))

    def _get_lang_code_str(self, name: str = None, lang_code: LangCode = None,
                           data_ml_json: dict = None) -> str:
        if lang_code is None and name is not None:
            lang_code_str = LangCode.detect_lang_code_str(name)
        elif lang_code is None and data_ml_json is not None:
            text = data_ml_json.get("title") or data_ml_json.get("name")
            lang_code_str = LangCode.detect_lang_code_str(text)
        elif lang_code is None:
            lang_code_str = self.user_context.get_effective_profile_preferred_lang_code_string()
        else:
            lang_code_str = lang_code.value
        return lang_code_str

    # The follwing methods may be irrelevant since GenericCRUD has equivalent methods
    # TODO: We may want to remove these following methods

    # This method returns the value in a selected column by a condition, the condition can be none or a value
    # for a specific column
    def select_one_by_value_with_none_option(
            self, condition_column_name, condition_column_value,
            view_table_name: str = None,
            select_column_name: str = None) -> Any:
        view_table_name = view_table_name or self.default_view_table_name
        # Returns the row id if select_column_name is None
        select_column_name = select_column_name or self.default_column
        """Selects a column from the table based on a WHERE clause and returns it as a list of dictionaries."""

        self.logger.warning("this method is deprecated, use select_one_tuple_by_id instead")
        validate_select_table_name(view_table_name)
        select_query = f"SELECT {select_column_name} " f"FROM {self.schema_name}.{view_table_name} " + (
            f"WHERE `{condition_column_name}` = \"{condition_column_value}\" " if (condition_column_name and
                                                                                   condition_column_value)
            else f"WHERE `{condition_column_name}` IS NULL ")

        self.cursor.execute(select_query)
        result = self.cursor.fetchall()
        # Extract the first element from the first tuple in the result
        result = result[0][0] if result else None

        return result

    def select_multi_tuple_by_value_with_none_option(self,
                                                     condition_column_name,
                                                     condition_column_value,
                                                     view_table_name: str = None, select_clause_value: str = "*",

                                                     limit: int = DEFAULT_SQL_SELECT_LIMIT,
                                                     order_by: str = None) -> list:
        view_table_name = view_table_name or self.default_view_table_name
        # Returns the row id if select_column_name is None
        select_column_name = f"`{select_clause_value}`" if select_clause_value != "*" else "*"
        """Selects a column from the table based on a WHERE clause and returns it as a list of dictionaries."""

        self.logger.warning("this method is deprecated, use select_multi_tuple_by_id instead")
        validate_select_table_name(view_table_name)
        select_query = f"SELECT {select_column_name} " f"FROM {self.schema_name}.{view_table_name} " + (
            f"WHERE `{condition_column_name}` = \"{condition_column_value}\" " if (condition_column_name and
                                                                                   condition_column_value)
            else f"WHERE {condition_column_name} IS NULL ") + (
            f"ORDER BY {order_by} " if order_by else "") + f"LIMIT {limit}"
        self.cursor.execute(select_query)
        result = self.cursor.fetchall()
        return result

    def select_multi_dict_by_value_with_none_option(self,
                                                    condition_column_name,
                                                    condition_column_value,
                                                    view_table_name: str = None, select_clause_value: str = "*",

                                                    limit: int = DEFAULT_SQL_SELECT_LIMIT,
                                                    order_by: str = None) -> list:
        self.logger.warning("this method is deprecated, use select_multi_tuple_by_id instead")
        """Selects a column from the table based on a WHERE clause and returns it as a list of dictionaries."""
        result = self.select_multi_tuple_by_value_with_none_option(condition_column_name, condition_column_value,

                                                                   view_table_name=view_table_name,
                                                                   select_clause_value=select_clause_value,
                                                                   limit=limit, order_by=order_by)
        return [self.convert_to_dict(row, select_clause_value) for row in result]



from logger_local.MetaLogger import MetaLogger

from .constants import LOGGER_MAPPING_CODE_OBJECT
from .generic_crud import GenericCRUD


class GenericMapping(GenericCRUD, metaclass=MetaLogger, object=LOGGER_MAPPING_CODE_OBJECT):
    def __init__(self, default_schema_name: str = None,
                 default_table_name: str = None,
                 default_view_table_name: str = None,
                 default_id_column_name: str = None,
                 default_entity_name1: str = None,
                 default_entity_name2: str = None,
                 is_test_data: bool = False):
        GenericCRUD.__init__(self, default_schema_name=default_schema_name,
                             default_table_name=default_table_name, default_view_table_name=default_view_table_name,
                             default_id_column_name=default_id_column_name, is_test_data=is_test_data)
        self.default_entity_name1 = default_entity_name1
        self.default_entity_name2 = default_entity_name2

    # We added the schema_name parameter to avoid using USE and void creating a new instance/object
    # of the database for mapping.
    def insert_mapping(self, *, entity_id1: int, entity_id2: int,
                       entity_name1: str = None, entity_name2: str = None,
                       schema_name: str = None, data_json: dict = None,
                       ignore_duplicate: bool = False) -> int:
        """Inserts a new link between two entities and returns the id of the
          new row or -1 if an error occurred.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param data_json: The data to insert.
        :param ignore_duplicate: If True, ignore duplicate rows.
        :return: The id of the new row or -1 if an error occurred.
        """

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.schema_name
        table_name = f"{entity_name1}_{entity_name2}_table"
        data_json = data_json or {}
        data_json[f"{entity_name1}_id"] = entity_id1
        data_json[f"{entity_name2}_id"] = entity_id2
        return self.insert(schema_name=schema_name, data_json=data_json, table_name=table_name,
                           ignore_duplicate=ignore_duplicate)

    # TODO: do we need delete_mapping_by_id?
    def delete_mapping_by_two_ids(self, *, entity_id1: int, entity_id2: int,
                                  entity_name1: str = None, entity_name2: str = None,
                                  schema_name: str = None) -> None:
        """ Deletes a link between two entities.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :return: None
        """

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.schema_name
        table_name = f"{entity_name1}_{entity_name2}_table"

        where = f"{entity_name1}_id=%s AND {entity_name2}_id=%s"
        params = (entity_id1, entity_id2)
        self.delete_by_where(schema_name=schema_name, table_name=table_name, where=where, params=params)

    def select_multi_mapping_tuple_by_id(self, *, entity_id1: int, entity_id2: int,
                                         entity_name1: str = None, entity_name2: str = None,
                                         schema_name: str = None, select_clause_value: str = "*") -> list:
        """Selects a row from the database by id.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param select_clause_value: The columns to select.
        :return: A list of dictionaries representing the rows.
        """
        assert entity_id1 is not None
        assert entity_id2 is not None

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.schema_name
        table_name = f"{entity_name1}_{entity_name2}_view"
        where = f"{entity_name1}_id=%s AND {entity_name2}_id=%s"
        params = (entity_id1, entity_id2)
        result = self.select_multi_tuple_by_where(schema_name=schema_name, view_table_name=table_name,
                                                  select_clause_value=select_clause_value,
                                                  where=where, params=params)
        return result

    def select_multi_mapping_dict_by_id(self, *, entity_id1: int, entity_id2: int,
                                        entity_name1: str = None, entity_name2: str = None,
                                        schema_name: str = None, select_clause_value: str = "*") -> list:
        """Selects a row from the database by id.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param select_clause_value: The columns to select.
        :return: A list of dictionaries representing the rows.
        """

        result = self.select_multi_mapping_tuple_by_id(entity_name1=entity_name1, entity_name2=entity_name2,
                                                       entity_id1=entity_id1, entity_id2=entity_id2,
                                                       schema_name=schema_name,
                                                       select_clause_value=select_clause_value)

        return [self.convert_to_dict(row, select_clause_value) for row in result]



from logger_local.LoggerComponentEnum import LoggerComponentEnum

DEFAULT_SQL_SELECT_LIMIT = 100

# connector / cursor
DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_ID = 112
DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_NAME = 'database_mysql_local\\connector'
CONNECTOR_DEVELOPER_EMAIL = 'idan.a@circ.zone'
LOGGER_CONNECTOR_CODE_OBJECT = {
    'component_id': DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': CONNECTOR_DEVELOPER_EMAIL
}
LOGGER_CONNECTOR_TEST_OBJECT = LOGGER_CONNECTOR_CODE_OBJECT.copy()
LOGGER_CONNECTOR_TEST_OBJECT['component_category'] = LoggerComponentEnum.ComponentCategory.Unit_Test.value

# generic_crud
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID = 206
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME = 'database_mysql_local\\generic_crud'
GENERIC_CRUD_DEVELOPER_EMAIL = 'akiva.s@circ.zone'
LOGGER_CRUD_CODE_OBJECT = {
    'component_id': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': GENERIC_CRUD_DEVELOPER_EMAIL
}
LOGGER_CRUD_TEST_OBJECT = LOGGER_CRUD_CODE_OBJECT.copy()
LOGGER_CRUD_TEST_OBJECT['component_category'] = LoggerComponentEnum.ComponentCategory.Unit_Test.value

# generic_crud_ml
DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_ID = 7001
DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_NAME = 'database_mysql_local\\generic_crud_ml'
GENERIC_CRUD_ML_DEVELOPER_EMAIL = 'tal.g@circ.zone'
LOGGER_CRUD_ML_CODE_OBJECT = {
    'component_id': DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': GENERIC_CRUD_ML_DEVELOPER_EMAIL
}
LOGGER_CRUD_ML_TEST_OBJECT = LOGGER_CRUD_ML_CODE_OBJECT.copy()
LOGGER_CRUD_ML_TEST_OBJECT['component_category'] = LoggerComponentEnum.ComponentCategory.Unit_Test.value

# generic_mapping
DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_ID = 7002
DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_NAME = 'database_mysql_local\\generic_mapping'
GENERIC_MAPPING_DEVELOPER_EMAIL = 'sahar.g@circ.zone'
LOGGER_MAPPING_CODE_OBJECT = {
    'component_id': DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': GENERIC_MAPPING_DEVELOPER_EMAIL
}

LOGGER_MAPPING_TEST_OBJECT = LOGGER_MAPPING_CODE_OBJECT.copy()
# alternative to the above, so we can update multiple keys at once
LOGGER_MAPPING_TEST_OBJECT.update(
    {'component_category': LoggerComponentEnum.ComponentCategory.Unit_Test.value}
)



from typing import Any

from logger_local.MetaLogger import MetaLogger

from .constants import LOGGER_CONNECTOR_CODE_OBJECT


class Cursor(metaclass=MetaLogger, object=LOGGER_CONNECTOR_CODE_OBJECT):
    def __init__(self, cursor) -> None:
        self.cursor = cursor

    # TODO: If environment <> prod1 and dvlp1 break down using 3rd party package and analyze the formatted_sql
    #  and call private method _validate_select_table_name(table_name)
    def execute(self, sql_statement: str, sql_parameters: tuple = None) -> None:
        # TODO: validate_select_table_name(table_name)
        # TODO: remove the old approach, including the try-except
        if sql_parameters:
            quoted_parameters = ["'" + str(param) + "'" for param in sql_parameters]
            formatted_sql = sql_statement % tuple(quoted_parameters)
            sql_parameters_str = ", ".join(quoted_parameters)
        else:
            formatted_sql = sql_statement
            sql_parameters_str = "None"
        self.logger.info('database-mysql-local cursor.py execute()', object={
            "formatted_sql": formatted_sql,
            "sql_parameters": sql_parameters_str,
            "sql_statement": sql_statement
        })
        self.cursor.execute(sql_statement, sql_parameters)

    def fetchall(self) -> Any:
        return self.cursor.fetchall()

    def fetchone(self) -> Any:
        return self.cursor.fetchone()

    def description(self) -> Any:
        return self.cursor.description

    def lastrowid(self) -> int:
        return self.cursor.lastrowid

    def close(self) -> None:
        self.cursor.close()
